@@ -1,23 +1,11 @@
 """
-SCIRIS
-
-.. autosummary::
-    :toctree: _autosummary
-
-    sc_asd
-    sc_dataframe
-    sc_fileio
-    sc_math
-    sc_odict
-    sc_parallel
-    sc_plotting
-    sc_utils
-    sc_version
+Sciris root module
 
+Functions in Sciris are of course organized into submodules. However, standard
+usage is to call the functions directly, e.g. ``sc.parallelize()`` instead
+of ``sc.sc_parallel.parallelize()``.
 """
 
-
-
 # Import everything
 from .sc_version   import *
 from .sc_utils     import *
